#pragma clang diagnostic push
#pragma ide diagnostic ignored "readability-make-member-function-const"
#pragma ide diagnostic ignored "readability-convert-member-functions-to-static"

#include <botan/argon2.h>
#include <botan/totp.h>
#include <jwt-cpp/traits/kazuho-picojson/traits.h>
#include <jwt-cpp/jwt.h>

#include "controllers/controllers.h"
#include "db/db.h"
#include "dto/dto.h"

namespace api {
    void auth::login(req_type req, cbk_type cbk) {
        Json::Value &json = *req->jsonObject();
        try {
            std::string username = dto::json_get<std::string>(json, "username").value();
            std::string password = dto::json_get<std::string>(json, "password").value();
            std::optional<std::string> otp = dto::json_get<std::string>(json, "otp");

            auto db = drogon::app().getDbClient();

            db::MapperUser user_mapper(db);
            auto db_users = user_mapper.findBy(
                    db::Criteria(db::User::Cols::_name, db::CompareOps::EQ, username) &&
                    db::Criteria(db::User::Cols::_gitlab, db::CompareOps::EQ, 0)
            );
            if (db_users.empty()) {
                cbk(dto::Responses::get_unauth_res("Invalid username or password"));
                return;
            }
            db::User &db_user = db_users.at(0);
            if (!Botan::argon2_check_pwhash(password.c_str(), password.size(), db_user.getValueOfPassword())) {
                cbk(dto::Responses::get_unauth_res("Invalid username or password"));
                return;
            }
            if (db::User_getEnumRole(db_user) == db::UserRole::DISABLED) {
                cbk(dto::Responses::get_unauth_res("Account is disabled"));
                return;
            }

            const auto tfa = db::User_getEnumTfaType(db_user);
            if (tfa != db::tfaTypes::NONE) {
                if (!otp.has_value()) {
                    if (tfa == db::tfaTypes::EMAIL) send_mail(db_user);
                    return cbk(dto::Responses::get_success_res());
                }
                if (!verify2fa(db_user, std::stoi(otp.value())))
                    return cbk(dto::Responses::get_unauth_res("Incorrect 2fa"));
            }

            cbk(dto::Responses::get_login_res(get_token(db_user)));
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }

    void auth::signup(req_type req, cbk_type cbk) {
        Json::Value &json = *req->jsonObject();
        try {
            std::string username = dto::json_get<std::string>(json, "username").value();
            std::string password = dto::json_get<std::string>(json, "password").value();

            db::MapperUser user_mapper(drogon::app().getDbClient());

            auto existing_users = user_mapper.count(
                    db::Criteria(db::User::Cols::_name, db::CompareOps::EQ, username) &&
                    db::Criteria(db::User::Cols::_gitlab, db::CompareOps::EQ, 0)
            );
            if (existing_users != 0) {
                cbk(dto::Responses::get_badreq_res("Username is already taken"));
                return;
            }

            //std::string hash = Botan::argon2_generate_pwhash(password.c_str(), password.size(), *rng, 1, 256*1024, 2);
            std::string hash = Botan::argon2_generate_pwhash(password.c_str(), password.size(), *rng, 1, 16*1024, 1);

            db::User new_user;
            new_user.setName(username);
            new_user.setPassword(hash);
            new_user.setGitlab(0);
            new_user.setRole(db::UserRole::DISABLED);
            new_user.setRootId(0);
            new_user.setTfaType(db::tfaTypes::NONE);

            user_mapper.insert(new_user);
            generate_root(new_user);
            cbk(dto::Responses::get_success_res());
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }

    void auth::refresh(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        db::Token token = dto::get_token(req);

        db::MapperToken token_mapper(drogon::app().getDbClient());
        token_mapper.deleteOne(token);
        cbk(dto::Responses::get_login_res( get_token(user)));
    }

    void auth::logout_all(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        revoke_all(user);
        cbk(dto::Responses::get_success_res());
    }

    void auth::change_password(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        Json::Value &json = *req->jsonObject();
        try {
            std::string old_pw = dto::json_get<std::string>(json, "oldPassword").value();
            std::string new_pw = dto::json_get<std::string>(json, "newPassword").value();

            auto db = drogon::app().getDbClient();
            db::MapperUser user_mapper(db);

            if (!Botan::argon2_check_pwhash(old_pw.c_str(), old_pw.size(), user.getValueOfPassword()))
                return cbk(dto::Responses::get_unauth_res("Old password is wrong"));

            std::string hash = Botan::argon2_generate_pwhash(new_pw.c_str(), new_pw.size(), *rng, 1, 256*1024, 2);

            user.setPassword(hash);
            user_mapper.update(user);
            revoke_all(user);
            cbk(dto::Responses::get_success_res());
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }
}

#pragma clang diagnostic pop